// SPDX-License-Identifier: BSD-3-Clause

#include <pthread.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdatomic.h>
#include "consumer.h"
#include "ring_buffer.h"
#include "packet.h"
#include "utils.h"

// _Atomic type variables to avoid race conditions for multiple threads
_Atomic unsigned long global_seq_num; // assign a number to each packet processed by any consumer
_Atomic unsigned long write_seq_num; // assign a number to each packet written to the log file

//mutex and condition variables for thread synchronization
pthread_mutex_t write_mutex; // protects acces to shared resources during write
pthread_cond_t write_condition; // used to signal between threads

void cleanup_thread_resources(void)
{
	// destroy mutex and condition variables
	pthread_mutex_destroy(&write_mutex);
	pthread_cond_destroy(&write_condition);
}


void write_to_file(int file_descriptor, unsigned long packet_sequence, char log_entry[], int length)
{
	pthread_mutex_lock(&write_mutex); // ensures only one thread writes at a time
	// wait until it's the thread's turn to write
	while (packet_sequence != write_seq_num)
		pthread_cond_wait(&write_condition, &write_mutex);

	// write to log_entry file and increment sequence
	write(file_descriptor, log_entry, length);
	// using atomic_fetch-add to make sure incrementing isn't interrupted by other threads
	atomic_fetch_add(&write_seq_num, 1);

	// signal the waiting threads
	pthread_cond_broadcast(&write_condition);
	pthread_mutex_unlock(&write_mutex);
}

void *consumer_thread(void *arg)
{
	so_consumer_ctx_t *ctx = (so_consumer_ctx_t *)arg; // for coding style purposes
	char packet_buffer[256]; // buffer for storing recieved packet data
	char log_entry[256]; // buffer for storing log entry
	unsigned long packet_sequence;
	int file_descriptor = ctx->log_file;

	while (true) {
		// dequeue packet from ring buffer
		pthread_mutex_lock(&write_mutex); // get exclusive access for a thread
		ssize_t received_size = ring_buffer_dequeue(ctx->producer_rb, packet_buffer, 256);

		if (received_size <= 0 && ctx->producer_rb->stop) {
			pthread_mutex_unlock(&write_mutex); // release mutex before exiting
			break;
		}
		// increment the global sequence number and release the mutex
		packet_sequence = atomic_fetch_add(&global_seq_num, 1); // makes sure incrementing isn't interrupted by other threads
		pthread_mutex_unlock(&write_mutex);

		// process packet and write to log file
		struct so_packet_t *packet = (struct so_packet_t *)packet_buffer;
		int action_result = process_packet(packet);
		unsigned long hash_value = packet_hash(packet);
		unsigned long time_stamp = packet->hdr.timestamp;

		int length = snprintf(log_entry, sizeof(log_entry), "%s %016lx %lu\n",
							action_result == PASS ? "PASS" : "DROP", hash_value, time_stamp);
		write_to_file(file_descriptor, packet_sequence, log_entry, length);
	}
	cleanup_thread_resources();
	return NULL;
}

void initialize_consumer_context(so_consumer_ctx_t *ctx, int file_descriptor, struct so_ring_buffer_t *rb)
{
	ctx->producer_rb = rb;
	ctx->log_file = file_descriptor;
	ctx->log_mutex = (pthread_mutex_t)PTHREAD_MUTEX_INITIALIZER;
}


int create_consumers(pthread_t *tids,
					int num_consumers,
					struct so_ring_buffer_t *rb,
					const char *out_filename)
{
	int file_descriptor = open(out_filename, O_RDWR | O_CREAT | O_APPEND, 0666);

	DIE(file_descriptor < 0, "Error opening output file");

	for (int i = 0; i < num_consumers; i++) {
		so_consumer_ctx_t *ctx = malloc(sizeof(so_consumer_ctx_t));// allocate memory for consumer context

		if (!ctx)
			DIE(1, "malloc failed");

		initialize_consumer_context(ctx, file_descriptor, rb);

		// create consumer threads
		pthread_create(&tids[i], NULL, consumer_thread, ctx);
	}

	return num_consumers;
}
